#!/usr/bin/env ruby -wKU
#
#  Sync-N-Link for FCP
#
#  Created by Gregory Clarke on 2008-08-02.
#  Copyright (c) 2008 Intelligent Assistance Software All rights reserved.
#

require 'osx/cocoa'
require 'synchronicity_algorithm.bundle'

class AppController < OSX::NSObject
	
	ib_outlets :mainWindow, :deactivateMenuItem, :saveMenuItem
	ib_outlets :syncNLinkView, :openFileButton, :xmlIcon, :xmlFileName, :binClipSequenceCount, :offsetAdjustmentField, :optionsTabView, :cameraAudioCheckbox,:trimRadiobuttons, :lognotesLabel, :lognotesPopup, :multiclipCheckbox, :multiclipAnglesPopup, :cameraAnglesLabel, :cameraAudioLabel, :audioTracksPopup, :saveFileButton, :openAutomaticallyCheckbox, :spinner
	ib_outlets :progressWindow, :progressView, :progressIndicator, :progressMessage
	ib_outlets :regoView, :deactivateAreaView, :webView, :webSpinner, :problemMessage, :closeRegoButton
	ib_outlets :preferencesWindow, :videoTimecodeRadiobuttons, :audioTimecodeRadiobuttons, :newBinsRadioButtons, :clipNamingRadioButtons, :clipNamingCheckbox
	
	def initialize
		# radio buttons default to first in the matrix; checkboxes default to off
		default_values = { "CameraAudio" => 1, "ClipNaming" => 1, "OpenAutomatically" => 1 }
		defaultsController = OSX::NSUserDefaultsController.sharedUserDefaultsController
		defaultsController.setInitialValues(default_values)
	end
	
	def awakeFromNib
		@xmlIcon.setHidden(true)
		@xmlIcon.setToolTip("")
		@xmlFileName.setStringValue("")
		@binClipSequenceCount.setStringValue("")
		enableBinClipOptions(false)
		enableSequenceOptions(false)
		# will use CoreAnimation for the window changes:
		@mainWindow.contentView.wantsLayer = true
		@webView.setDrawsBackground(false)
	end
	
	def enableBinClipOptions(bool)
		@cameraAudioCheckbox.setEnabled(bool)
		@trimRadiobuttons.setEnabled(bool)
		@lognotesPopup.setEnabled(bool)
		@multiclipCheckbox.setEnabled(bool)
		@multiclipAnglesPopup.setEnabled(bool)
		if bool==true
			@lognotesLabel.setTextColor(OSX::NSColor.blackColor)
			@cameraAnglesLabel.setTextColor(OSX::NSColor.blackColor)
		else
			@lognotesLabel.setTextColor(OSX::NSColor.disabledControlTextColor) # so it looks like it is disabled
			@cameraAnglesLabel.setTextColor(OSX::NSColor.disabledControlTextColor)
		end
	end
	
	def enableSequenceOptions(bool)
		if bool==true
			@cameraAudioLabel.setTextColor(OSX::NSColor.blackColor)
			@audioTracksPopup.addItemWithTitle("Track A1 only")
			(@number_of_audio_tracks_in_sequence-1).times do |i|
				@audioTracksPopup.addItemWithTitle("Tracks A1 to A#{i+2}")
			end
		else
			@cameraAudioLabel.setTextColor(OSX::NSColor.disabledControlTextColor)
			@audioTracksPopup.removeAllItems
		end
		@audioTracksPopup.setEnabled(bool)
	end

	### rego methods ###

	def applicationWillFinishLaunching(sender) # applicationWillFinishLaunching: is sent before application:openFile: (so it can deal with a file dropped onto the app icon)
		#puts "applicationWillFinishLaunching"
		@saveFileButton.setEnabled(false)
		@saveMenuItem.setEnabled(false)
		check_for_expired_beta()
		if @this_rego[0]["active"]
			@deactivateMenuItem.setEnabled(true)
			showSyncNLinkView()
		elsif @this_rego[0]["mac_problem"]==true
			check_for_rego(@this_rego)
		else
			@deactivateMenuItem.setEnabled(false)
			displayDemoSheet()
		end
	end

	def displayDemoSheet
		alert = OSX::NSAlert.alloc.init
		alert.setMessageText("#{SKU} has not been activated for this computer.")
		alert.setAlertStyle(OSX::NSInformationalAlertStyle)
		alert.setInformativeText("Would you like to activate #{SKU}, or try it in demo mode (which syncs up to 10 random clips)?")
		alert.addButtonWithTitle("Activate")
		alert.addButtonWithTitle("Buy")
		alert.addButtonWithTitle("Demo")
		alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, "didEndDemoSheet:returnCode:contextInfo:", nil)
	end

	def didEndDemoSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		sheet.window.orderOut(nil)
		if buttonClicked==OSX::NSAlertSecondButtonReturn
			system('open', 'http://www.AssistedEditing.com/Sync-N-Link/buy')
		end
		if buttonClicked==OSX::NSAlertThirdButtonReturn
			showSyncNLinkView()
		else
			newFrame = self.newFrameForNewContentView(@regoView)
			showActivationView(newFrame, WEB_APP + "license/" + SKU.gsub(" ", "%20") + "?sn=" + @this_rego[0]["number"].to_s + "&mac=" + MAC) # NEW
		end
	end

	def showDeactivation(sender)
		if @deactivateMenuItem.title=="Activate..."
			newFrame = self.newFrameForNewContentView(@regoView)
			showActivationView(newFrame, WEB_APP + "license/" + SKU.gsub(" ", "%20") + "?sn=" + @this_rego[0]["number"].to_s + "&mac=" + MAC) # NEW
		else
			newFrame = self.newFrameForNewContentView(@deactivateAreaView) # 500 x 200
			@mainWindow.setMinSize(OSX::NSSize.new(500, 200+22))
			@mainWindow.setMaxSize(OSX::NSSize.new(500, 200+22))
			@mainWindow.setContentView(@deactivateAreaView)
			showActivationView(newFrame, WEB_APP + "destroy/" + @this_rego[0]["number"].to_s)
		end
		@closeRegoButton.setTitle("Cancel")
	end
	ib_action :showDeactivation # Sync-N-Link > Deactivate
	
	def showActivationView(frame, url)
		#puts "showActivationView #{url}"
		@mainWindow.setFrame_display frame, true
		@mainWindow.setContentView(@regoView)
		@mainWindow.setTitle("Activation")
		# notification
		center = OSX::NSNotificationCenter.defaultCenter
		center.addObserver_selector_name_object(self, :webProgress, nil, @webView)
		@webView.mainFrame.loadRequest( OSX::NSURLRequest.requestWithURL( OSX::NSURL.URLWithString(url) ) )
	end
	
	def webProgress(notification)
		case notification.name
		when "WebProgressStartedNotification"
			@webSpinner.startAnimation(self)
		when "WebProgressFinishedNotification"
			@webSpinner.stopAnimation(self)
			if @webView.mainFrame.dataSource.nil? # in case page doesn't load...
				@webView.mainFrame.stopLoading
				@problemMessage.setHidden(false)
				@webView.setHidden(true)
			else
				get_web_app_response(@webView.mainFrame.dataSource.response.URL)
			end
		end
	end 
	
	def closeActivationView(sender)
		if @closeRegoButton.title=="Continue" || @closeRegoButton.title=="Cancel"
			showSyncNLinkView()
		else
			self.appTerminate
		end
	end
	ib_action :closeActivationView # on click of Registration window's Continue button

	### end rego methods ###

	def showSyncNLinkView
		#puts "showSyncNLinkView"
		setupView(@syncNLinkView, SKU, 430, 530)
	end

	def setupView(view, title, width, height)
		@mainWindow.setTitle(title)
		oldFrame = @mainWindow.frame
		newFrame = self.newFrameForNewContentView(view)
		@mainWindow.setMinSize(OSX::NSSize.new(width, height+22))
		@mainWindow.setMaxSize(OSX::NSSize.new(width, height+22))
		@mainWindow.setFrame_display_animate(newFrame, true, true)
		@mainWindow.setContentView(view)
	end

	def newFrameForNewContentView(view)
		newFrameRect = @mainWindow.frameRectForContentRect(view.frame)
		oldFrameRect = @mainWindow.frame
		newSize = newFrameRect.size
		oldSize = oldFrameRect.size
		frame = @mainWindow.frame
		frame.size = newSize
		frame.origin.y = frame.origin.y - (newSize.height - oldSize.height)
		frame
	end
	
	def showHelp(sender)
		openPDF("Sync-N-Link for FCP - Help.pdf")
	end
	ib_action :showHelp # Help > Sync-N-Link Help
	
	def openPDF(filename)
		path_to_file = __FILE__.sub("AppController.rb",filename)
		system('open', path_to_file)
	end

	def visitWebsite(sender)
		system('open', 'http://www.AssistedEditing.com/')
	end
	ib_action :visitWebsite # Help > Visit www.AssistedEditing.com

	def sendFeedback(sender)
		OSX::JRFeedbackController.showFeedback
		# sendSupportEmail("problem")
	end
	ib_action :sendFeedback # Sync-N-Link > Provide Feedback

	def sendSupportEmail(subject)
		system('open', "mailto:techsupport@intelligentassistance.com?subject=Sync-N-Link #{subject}&body=Attached is the XML file that Sync-N-Link for FCP was unable to process.")
	end

	def chooseXmlFile(sender)
		filetypes = ["xml"]
		oPanel = OSX::NSOpenPanel.openPanel
		oPanel.setAllowsMultipleSelection(false)
		oPanel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, filetypes, @mainWindow, self, 'didEndOpenSheet:returnCode:contextInfo:', nil)
	end
	ib_action :chooseXmlFile # File > Open, or click Open button

	def didEndOpenSheet_returnCode_contextInfo(oPanel, buttonClicked, info)
		@mainWindow.setTitle(SKU)
		if buttonClicked == OSX::NSOKButton
			oPanel.orderOut(self)
			file = oPanel.URL.path.to_s
			application_openFile(self, file)
		else
			@xmlIcon.setHidden(true)
			@xmlIcon.setToolTip("")
			@xmlFileName.setStringValue("")
			@binClipSequenceCount.setStringValue("")
			@saveFileButton.setEnabled(false)
			@saveMenuItem.setEnabled(false)
			@openFileButton.setEnabled(true)
		end
	end

	def application_openFile(sender, path)
		# add this file to the Recent menu
		OSX::NSDocumentController.sharedDocumentController().noteNewRecentDocumentURL(OSX::NSURL.fileURLWithPath(path))
		begin
			@input_bin_file = path.to_s
			@input_bin_file_name = File.basename(@input_bin_file, ".xml")
			@progressMessage.setStringValue("Opening \"" + @input_bin_file_name + "\"...")
			@progressIndicator.setIndeterminate(true)
			@progressIndicator.setUsesThreadedAnimation(true)
			@progressIndicator.startAnimation(self)
			displayProgressSheet()
			parse_bins_xml(path)
			@xmlIcon.setHidden(false)
			@xmlIcon.setToolTip(path.to_s)
			@xmlFileName.setStringValue(@input_bin_file_name + ".xml")
		rescue
			didEndProgressSheet_returnCode_contextInfo(@progressWindow, -1, nil)
			alert = OSX::NSAlert.alloc.init
			alert.setMessageText("There was a problem with this XML file.")
			alert.setAlertStyle(OSX::NSInformationalAlertStyle)
			#puts "RESCUE!!! Problem with parse_bins_xml()"
			alert.setInformativeText("Please choose another XML file exported from\nFinal Cut Pro.\n\nWould you like to email this XML file to Tech Support?")
			alert.addButtonWithTitle("Email Tech Support")
			alert.addButtonWithTitle("Ignore")
			buttonClicked = alert.runModal
			sendSupportEmail("problem reading XML file") if buttonClicked==OSX::NSAlertFirstButtonReturn
			chooseXmlFile(nil)
		end
	end

	def displayProgressSheet
		@saveFileButton.setEnabled(false)
		@saveMenuItem.setEnabled(false)
		@syncNLinkView.displayIfNeeded
		self.appBeginSheet_forWindow(@progressWindow, @mainWindow)
		@progressView.displayIfNeeded
	end

	def didEndProgressSheet_returnCode_contextInfo(sheet, code, info)
		#puts "didEndProgressSheet_returnCode_contextInfo code #{code}"
		sheet.orderOut(nil) # or self?
		@progressMessage.setStringValue("")
		@progressIndicator.stopAnimation(self)
		self.appEndSheet(sheet)
		if code==1 # we've finished parsing the XML file
			#puts "PARSER"
			# enable the interface elements
			@offsetAdjustmentField.setEnabled(true)
			@offsetAdjustmentField.setStringValue("0")
			if @has_sequence_with_clips==false # BINS
				#puts "BINS"
				enableBinClipOptions(true)
				enableSequenceOptions(false)
				if @has_camera_audio
					@cameraAudioCheckbox.setEnabled(true)
				else
					@cameraAudioCheckbox.setEnabled(false)
					@cameraAudioCheckbox.setState(0)
				end
				@optionsTabView.selectFirstTabViewItem(self)
				@saveFileButton.setTitle("Sync Bin Clips...")
			else # SEQUENCES
				#puts "SEQUENCE"
				enableSequenceOptions(true)
				enableBinClipOptions(false)
				@cameraAudioCheckbox.setState(1)
				@multiclipCheckbox.setState(0)
				@optionsTabView.selectLastTabViewItem(self)
				@saveFileButton.setTitle("Sync Sequence Clips...")
			end
		elsif code==-2 # detected multiclips in a sequence
			#puts "MULTICLIPS in #{info}"
			alert = OSX::NSAlert.alloc.init
			alert.setMessageText("Cannot replace audio in Multiclips")
			alert.setAlertStyle(OSX::NSInformationalAlertStyle)
			alert.setInformativeText("There are multiclips in the sequence \"#{info}\".\n\nCollapse the multiclips in Final Cut Pro using\nModify > Collapse Multiclip(s) and try again.")
			alert.runModal
		elsif code==2 # we've finished syncing video and audio
			#puts "SYNCER"
			#puts "@report[:message]:\n #{@report[:message]}"
				displayReportSheet()
		end
		if code < 0 # error condition
			@saveFileButton.setEnabled(false)
			@saveMenuItem.setEnabled(false)
		else
			@saveFileButton.setEnabled(true) # can save again
			@saveMenuItem.setEnabled(true)
		end
	end

	def startSyncNLink(sender)
		begin
			@progressMessage.setStringValue("Preparing...")
			@progressIndicator.setIndeterminate(true)
			@progressIndicator.startAnimation(self)
			displayProgressSheet()
			ready_to_run = check_for_timecode("xmeml") # have already confirmed it's xmeml
			@offset_adjustment = (@offsetAdjustmentField.stringValue.to_s).to_i
			#puts "@offset_adjustment = #{@offset_adjustment}"
			if @has_sequence_with_clips==false && ready_to_run # BINS
				@camera_audio_choice = case @cameraAudioCheckbox.state
					when 0 then "no" 
					when 1 then "yes"
				end
				@trim_preference = case @trimRadiobuttons.selectedCell.tag
					when 0 then "full duration"
					when 1 then "common duration"
					when 2 then "no trim"
				end
				@lognotes_choice = @lognotesPopup.titleOfSelectedItem
				@make_multiclips = case @multiclipCheckbox.state
					when 0 then false
					when 1 then true
				end
				@number_of_multiclip_angles = @multiclipAnglesPopup.titleOfSelectedItem.to_i
				@add_to_new_bins_pref = case @newBinsRadioButtons.selectedCell.tag
					when 0 then "synced"
					when 1 then "all"
				end
				@use_video_clip_name = case @clipNamingRadioButtons.selectedCell.tag
					when 0 then true
					when 1 then false
				end
				@add_synced_to_clip_name = case @clipNamingCheckbox.state
					when 0 then false
					when 1 then true
				end
				time1 = Time.now
				sync_n_link_bins()
				time2 = Time.now
				#puts "********** #{time2 - time1} seconds"
			elsif @has_sequence_with_clips && ready_to_run # SEQUENCES
				# work out audio tracks being kept
				@camera_audio_tracks = @audioTracksPopup.indexOfSelectedItem + 1
				#puts "@camera_audio_tracks = #{@camera_audio_tracks}"
				sync_n_link_sequences()
			end
		rescue
			puts "RESCUE!!! Problem with sync_n_link_bins()"
			didEndProgressSheet_returnCode_contextInfo(@progressWindow, -1, nil)
			alert = OSX::NSAlert.alloc.init
			alert.setMessageText("There was a problem with this XML file")
			alert.setAlertStyle(OSX::NSInformationalAlertStyle)
			alert.setInformativeText("Please choose another XML file exported from\nFinal Cut Pro.\n\nWould you like to email this XML file to Tech Support?")
			alert.addButtonWithTitle("Email Tech Support")
			alert.addButtonWithTitle("Ignore")
			buttonClicked = alert.runModal
			sendSupportEmail("problem syncing clips from XML file") if buttonClicked==OSX::NSAlertFirstButtonReturn
			chooseXmlFile(nil)
		end
	end
	ib_action :startSyncNLink # on click of Save XML button

	def displayTimecodeAlert(message, informative, showprefs)
		alert = OSX::NSAlert.alloc.init
		alert.setMessageText(message)
		alert.setInformativeText(informative)
		alert.setAlertStyle(OSX::NSCriticalAlertStyle)
		alert.addButtonWithTitle("Open XML...")
		alert.addButtonWithTitle("Show Preferences...") unless showprefs==false
		buttonClicked = alert.runModal
		if buttonClicked==OSX::NSAlertFirstButtonReturn
			chooseXmlFile(nil)
		else
			@preferencesWindow.makeKeyAndOrderFront(nil)
		end
	end

	def displayReportSheet
		report = OSX::NSAlert.alloc.init
		report.setMessageText("\"" + @input_bin_file_name + "\" report:")
		if @report[:fails]==0
			report.setAlertStyle(OSX::NSInformationalAlertStyle)
		else
			report.setAlertStyle(OSX::NSCriticalAlertStyle)
		end
		report.setInformativeText(@report[:message])
		button1 = report.addButtonWithTitle("Save XML File")
		button3 = report.addButtonWithTitle("Save Sync Report")
		# button2 = report.addButtonWithTitle("Save ALE File") if @has_sequence_with_clips==false && OSX::NSBundle.mainBundle.objectForInfoDictionaryKey("CFBundleVersion").to_s >= "1500"
		report.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, "didEndReportSheet:returnCode:contextInfo:", nil)
	end

	def didEndReportSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		# If you want to dismiss the sheet from within the alertDidEndSelector method before the modal delegate carries out an action in response to the return value, send orderOut: (NSWindow) to the window object obtained by sending window to the alert argument. This allows you to chain sheets, for example, by dismissing one sheet before showing the next from within the alertDidEndSelector method. Note that you should be careful not to call orderOut: on the sheet from elsewhere in your program before the alertDidEndSelector method is invoked.
		sheet.window.orderOut(nil)
		case buttonClicked
		when OSX::NSAlertFirstButtonReturn
			saveXmlFile()
		when OSX::NSAlertSecondButtonReturn
			saveReportFile()
		when OSX::NSAlertThirdButtonReturn
			saveALE()
		end
	end

	def saveReportFile
		sPanel = OSX::NSSavePanel.savePanel
		@mainWindow.setTitle("Save the Sync-N-Link report:")
		sPanel.setRequiredFileType("txt")
		sPanel.setExtensionHidden(false)
		sPanel.setCanSelectHiddenExtension(false)
		sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, @input_bin_file_name + " - Sync-N-Link Report", @mainWindow, self, 'didEndSaveReport:returnCode:contextInfo:', nil)
	end

	def didEndSaveReport_returnCode_contextInfo(sPanel, buttonClicked, info)
		@mainWindow.setTitle(SKU)
		sPanel.orderOut(self)
		if buttonClicked == OSX::NSOKButton
			@spinner.startAnimation(self)
			output_file = sPanel.URL.path.to_s
			begin
				### Save the Report ###
				fileManager = OSX::NSFileManager.defaultManager
				fileManager.createFileAtPath_contents_attributes(output_file, @report[:text], nil)
				@openFileButton.setEnabled(true)
				# system('open', output_file)
			rescue
				puts "RESCUE!!! Problem creating the report file"
				alert = OSX::NSAlert.alloc.init
				alert.setMessageText("Sorry!")
				alert.setInformativeText("There was a problem saving this report file.\n")
				alert.setAlertStyle(OSX::NSInformationalAlertStyle)
				buttonClicked = alert.runModal
			end
			@spinner.stopAnimation(self)
		end
		displayReportSheet()
	end

	def saveALE
		sPanel = OSX::NSSavePanel.savePanel
		@mainWindow.setTitle("Save the ALE file:")
		sPanel.setRequiredFileType("txt")
		sPanel.setExtensionHidden(false)
		sPanel.setCanSelectHiddenExtension(false)
		sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, @input_bin_file_name + " - ALE", @mainWindow, self, 'didEndSaveALE:returnCode:contextInfo:', nil)
	end

	def didEndSaveALE_returnCode_contextInfo(sPanel, buttonClicked, info)
		@mainWindow.setTitle(SKU)
		sPanel.orderOut(self)
		if buttonClicked == OSX::NSOKButton
			@spinner.startAnimation(self)
			output_file = sPanel.URL.path.to_s
			begin
				### Save the Report ###
				fileManager = OSX::NSFileManager.defaultManager
				fileManager.createFileAtPath_contents_attributes(output_file, @report[:ale], nil)
				@openFileButton.setEnabled(true)
				# system('open', output_file)
			rescue
				puts "RESCUE!!! Problem creating the ALE file"
				alert = OSX::NSAlert.alloc.init
				alert.setMessageText("Sorry!")
				alert.setInformativeText("There was a problem saving this ALE file.\n")
				alert.setAlertStyle(OSX::NSInformationalAlertStyle)
				buttonClicked = alert.runModal
			end
			@spinner.stopAnimation(self)
		end
		displayReportSheet()
	end

	def saveXmlFile
		sPanel = OSX::NSSavePanel.savePanel
		@mainWindow.setTitle("Save the new XML file:")
		sPanel.setRequiredFileType("xml")
		sPanel.setExtensionHidden(true)
		sPanel.setCanSelectHiddenExtension(false)
		sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, @input_bin_file_name + " - synced", @mainWindow, self, 'didEndSaveSheet:returnCode:contextInfo:', nil)
	end

	def didEndSaveSheet_returnCode_contextInfo(sPanel, buttonClicked, info)
		@mainWindow.setTitle(SKU)
		sPanel.orderOut(self)
		if buttonClicked == OSX::NSOKButton
			@spinner.startAnimation(self)
			@output_file = sPanel.URL.path.to_s
			begin
				### Save the XML ###
				save_this_file = File.new(@output_file, "w")
				save_this_file.puts @new_xml.sub("xmeml SYSTEM", "xmeml") # NOTE fix another weird problem
				save_this_file.close
				fileManager = OSX::NSFileManager.defaultManager
				fileAttributes = { :NSFileExtensionHidden => 1, :NSFileHFSCreatorCode => 1264941383 } # extension is hidden, Final Cut Pro is creator code
				fileManager.setAttributes_ofItemAtPath_error(fileAttributes, @output_file, nil)
				@openFileButton.setEnabled(true)
				@spinner.stopAnimation(self)
				if @openAutomaticallyCheckbox.state==1
					open_it = OSX::NSWorkspace.sharedWorkspace.openFile_withApplication(@output_file, "Final Cut Pro")
					if open_it==false
						@openAutomaticallyCheckbox.setState(0)
					end
				end
				# save multiclips bin if necessary
				# open multiclips bin if saved
			rescue
				puts "RESCUE!!! Problem creating the XML file"
				@spinner.stopAnimation(self)
				alert = OSX::NSAlert.alloc.init
				alert.setMessageText("Sorry!")
				alert.setInformativeText("There was a problem saving this XML file.\nPlease try again using a different exported file.")
				alert.setAlertStyle(OSX::NSInformationalAlertStyle)
				buttonClicked = alert.runModal
			end
		else
			@openFileButton.setEnabled(true)
		end
	end

	def applicationShouldTerminateAfterLastWindowClosed(sender)
		true
	end

end
